<!DOCTYPE html>
<html>
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width">
        <title>Spinta</title>
        <style>

        body {
            margin: 0;
            padding: 0;
        }

        .container {
            padding: 8px;
        }

        /* Table style generated using https://divtable.com/table-styler/ */
        table.table {
            width: 100%;
            text-align: left;
            border-collapse: collapse;
        }
        table.table td {
            border-bottom: 1px solid #b8b8b8;
        }
        table.table th {
            background: #E8E8E8;
            border-bottom: 3px solid #b8b8b8;
        }
        table.table td, table.table th {
            padding: 5px 4px;
        }
        table.table tbody td {
            font-size: 13px;
        }
        table.table thead th {
            font-size: 15px;
            font-weight: bold;
            color: #000000;
            text-align: left;
        }
        table.table tfoot td {
            font-size: 14px;
        }
        div.formats {
            text-align: right;
        }
        .warning {
            background-color: #ffe9ea;
            color: #b42b3f;
            margin: 0;
            padding: 8px 16px;
            font-size: 12px;
        }
        .warning a {
            color: #b42b3f;
        }
        </style>
    </head>
    <body>
        <div class="warning">
            <strong>Dėmesio!</strong>
            <a href="https://data.gov.lt/page/saugykla" target="_blank">Atvirų
            duomenų saugykla</a> šiuo metu yra
            <a href="https://atviriduomenys.readthedocs.io/api/index.html#statusas-ir-planas" target="_blank">aktyviai vystoma</a>,
            todėl galite susidurti su įvairaus pobūdžio sutrikimais, apie
            kuriuos maloniai prašome <a href="https://github.com/atviriduomenys/spinta/issues/new" target="_blank">pranešti</a>.
        </div>

        <div class="container">
            {% block content %}{% endblock %}
        </div>
    </body>
</html>
